@@ -15,12 +15,18 @@ pub struct PartialReciprocal {
 }
 
 /// A `Reciprocal` represents an integer division by any non-zero
-/// `u64`.  It consists of a regular `PartialReciprocal`, and a
-/// hardcoded result for `u64::MAX / d`.
+/// `u64`.  It replaces `PartialReciprocal`'s expression,
+///   `f(x) = (x + increment) * multiplier >> (64 + shift)`, where
+/// the inner addition is a saturating add by 0 or 1, with
+///   `g(x) = (x * multiplier + summand) >> (64 + shift)`, where
+/// both the multiplication and additions are in full 128 bit
+/// arithmetic.  This additional work lets us handle all the cases,
+/// including divisions by 1 and by `u64::MAX`.
 #[derive(Copy, Clone, Debug, PartialEq, Eq, Hash)]
 pub struct Reciprocal {
-    u64_max_result: u64,
-    base: PartialReciprocal,
+    multiplier: u64,
+    summand: u64, // Always 0 (round up) or `multiplier` (round down)
+    shift: u8,
 }
 
 impl PartialReciprocal {
@@ -111,14 +117,57 @@ impl PartialReciprocal {
 
     /// Computes `x / d`, where `d` is the divison for which this
     /// reciprocal was constructed.
+    ///
+    /// The library tries to dispatch to a reasonable implementation
+    /// for each platform.
     #[inline]
     #[must_use]
+    #[cfg(target_arch = "x86_64")]
     pub fn apply(self, x: u64) -> u64 {
+        self.apply_overflowing(x)
+    }
+
+    /// Computes `x / d`, where `d` is the divison for which this
+    /// reciprocal was constructed.
+    ///
+    /// The library tries to dispatch to a reasonable implementation
+    /// for each platform.
+    #[inline]
+    #[must_use]
+    #[cfg(not(target_arch = "x86_64"))]
+    pub fn apply(self, x: u64) -> u64 {
+        self.apply_saturating(x)
+    }
+
+    /// Computes `x / d` with saturating arithmetic.
+    #[inline]
+    #[must_use]
+    pub fn apply_saturating(self, x: u64) -> u64 {
         let shifted = x.saturating_add(self.increment as u64);
         let hi = ((self.multiplier as u128 * shifted as u128) >> 64) as u64;
 
         hi >> self.shift
     }
+
+    /// Computes `x / d` with an alternative saturating increment that
+    /// is shorter and equally efficient on x86-64.
+    #[inline]
+    #[must_use]
+    pub fn apply_overflowing(self, x: u64) -> u64 {
+        // Manually implement a saturating increment: we know
+        // `increment` is only 0 or 1, so we can recover from any
+        // overflow by subtracting 1 from `shifted`.  We can then
+        // expect LLVM to implement that as a subtract-with-borrow on
+        // x86-64.
+        let (mut shifted, overflow) = x.overflowing_add(self.increment as u64);
+        if overflow {
+            shifted = shifted.wrapping_sub(1);
+        }
+
+        let hi = ((self.multiplier as u128 * shifted as u128) >> 64) as u64;
+
+        hi >> self.shift
+    }
 }
 
 impl Reciprocal {
@@ -130,42 +179,53 @@ impl Reciprocal {
             return None;
         }
 
-        let u64_max_result = u64::MAX / d;
         if let Some(base) = PartialReciprocal::new(d) {
             return Some(Reciprocal {
-                u64_max_result,
-                base,
+                multiplier: base.multiplier,
+                summand: base.multiplier * base.increment as u64,
+                shift: base.shift,
             });
         }
 
-        // The two special cases below set `base` correctly for any `x
-        // < u64::MAX`, and all set `increment = 1`.
-        //
+        // The two special cases below work in 64.64 fixed point.
+        // While they have the usual `summand = multiplier` structure,
+        // their results differ when dividing u64::MAX and (u64::MAX -
+        // 1).  That's why we can't use the same constants for the
+        // `PartialReciprocal` sequence, which
+        // pre-saturating-increments the dividend, instead of actually
+        // adding the summand to the intermediate 128-bit product.
+
         // We can thus determine whether to use `u64_max_result` by
         // checking if the increment overflows.
         assert!(d == 1 || d == u64::MAX);
         if d == 1 {
-            // We can spoof the identity by adding 1 and scaling
-            // by u64::MAX / 2**64.
+            // We want to divide by 1, i.e., multiply by 1.
+            //
+            // We can fake it by scaling by by u64::MAX / 2**64 (1 -
+            // 2**-64), and adding (1 - 2**-64) again.
+            //
+            // It would even be correct to add 1 (x * u64::MAX / 2**64
+            // truncates to x - 1 for all u64 values), but that
+            // doesn't fit our pattern... and adding (1 - 2**-64)
+            // suffices to compensate even the worst-case
+            // approximation error (max x * [1 - (1 - 2**-64)]
+            // = max x / 2**64 = u64::MAX / 2**-64).
             return Some(Reciprocal {
-                u64_max_result,
-                base: PartialReciprocal {
-                    multiplier: u64::MAX,
-                    shift: 0,
-                    increment: 1,
-                },
+                multiplier: u64::MAX,
+                summand: u64::MAX,
+                shift: 0,
             });
         }
 
-        // And we can fake a division by u64::MAX with
-        // a multiplication by zero.
+        // And we can fake a division by u64::MAX with a
+        // multiplication by 2**-64, followed by adding 2**-64 again..
+        //
+        // For any value less than u64::MAX, the result is less than
+        // 1, so truncates to 0.  For u64::MAX, we get exactly 1.
         Some(Reciprocal {
-            u64_max_result,
-            base: PartialReciprocal {
-                multiplier: 0,
-                shift: 0,
-                increment: 1,
-            },
+            multiplier: 1,
+            summand: 1,
+            shift: 0,
         })
     }
 
@@ -174,27 +234,10 @@ impl Reciprocal {
     #[inline]
     #[must_use]
     pub fn apply(&self, x: u64) -> u64 {
-        let (shifted, wrapped) = x.overflowing_add(self.base.increment as u64);
+        let mut product = x as u128 * self.multiplier as u128;
+        product += self.summand as u128;
 
-        // This condition would ideally be `unlikely`.  We could also
-        // recover branch-freedom if we could convince llvm to emit
-        // conditional moves for something like
-        //
-        //     let fixup = if wrapped { self.u64_max_result } else { 0 }
-        //     let hi = ...;
-        //     (hi >> self.base.shift) + fixup
-        //
-        // The sequence above works because when the addition wraps,
-        // `shifted == 0`, so `hi == 0`.
-        if wrapped {
-            // `self.base.increment <= 1`, so the addition can only
-            // wrap if `x == u64::MAX`, and we already have the result
-            // for that division.
-            return self.u64_max_result;
-        }
-
-        let hi = ((self.base.multiplier as u128 * shifted as u128) >> 64) as u64;
-        hi >> self.base.shift
+        (product >> 64) as u64 >> self.shift
     }
 }
 
@@ -209,11 +252,15 @@ mod tests {
         let probe = |x: u64| {
             let expected = x / d;
             if let Some(p) = partial {
-                assert_eq!(p.apply(x), expected, "d={}, x={}", d, x);
+                assert_eq!(p.apply_saturating(x), expected, "d={}, x={}", d, x);
+                assert_eq!(p.apply_overflowing(x), expected, "d={}, x={}", d, x);
             }
 
             if let Some(r) = recip {
                 assert_eq!(r.apply(x), expected, "d={}, x={}", d, x);
+
+                #[cfg(feature = "nightly")]
+                assert_eq!(r.apply_branchfree(x), expected, "d={}, x={}", d, x);
             }
         };
 
